# Copyright (c) 2019 Capital Fund Management
# SPDX-License-Identifier: MIT

from .table import Tables
from .dataframe_table import DataFrameTable
from .function import Function, Functions
from .__version__ import __version__

__all__ = [Tables, DataFrameTable, Functions, Function, __version__]



import json


class Function:
    def __init__(self, alias, method):
        self.alias = alias
        self.method = method

    def __call__(self, *args, **kwargs):
        return self.method(*args)


class Functions:
    """
    Function manager exposed as a dictionary to keep track of all registered functions and create a
    combined schema.
    """

    def __init__(self, *args):
        self.functions = {}

    def __setitem__(self, key, value):
        assert isinstance(value, Function)
        self.functions[key.lower()] = value

    def __getitem__(self, key):
        return self.functions[key.lower()]

    def render_evaluate(self, request, do_print=True):
        """
        # POST /evaluate
        tables.render_evaluate()
        """
        jreq = json.loads(request)

        if 'function' not in jreq['body']:
            raise ValueError(f"<function> field is expected to evaluate method -> {jreq['body']}")

        function_name = jreq['body']['function']

        if 'data' not in jreq['body']:
            raise ValueError(f"<data> field is expected to evaluate method -> {jreq['body']}")

        arg_dict = jreq['body']['data']

        function = self[function_name.lower()]
        ret_value = list()

        arg_dict_values = [value for key, value in sorted(arg_dict.items())]

        for arg_values in zip(*arg_dict_values):
            try:
                invoke_result = function(*arg_values)
                ret_value.append(invoke_result)
            except Exception as e:
                ret_value.append(str(e))

        json_return = json.dumps(ret_value)

        if do_print:
            print(json_return)
        else:
            return json_return



import re
import unicodedata
import pandas as pd
from collections import Counter

from .table import BaseTable


class DataFrameTable(BaseTable):
    """
    This class represents a jupytab-ready table that exposes a Pandas DataFrame.
    """

    def __init__(self, alias, dataframe=None, refresh_method=None, include_index=False):
        """
        alias -- Descriptive name of the table, that will be displayed in Tableau.

        dataframe -- Pandas DataFrame to be accessed from Tableau (may be None if a
        refresh_method is provided).

        refresh_method -- Optional method callback that will be called every time
        Tableau needs to access the data (for instance when the DataSource is refreshed).
        It takes no argument and must return a DataFrame with the same column layout
        (schema) as the original DataFrame (if any).

        include_index -- Add Index as column(s) in the output data to Tableau.
        """
        BaseTable.__init__(self, alias=alias)

        self._dataframe = dataframe
        self._refresh_method = refresh_method
        self._include_index = include_index
        self._index_separator = '_'

        self.types_mapping = {
            'object': 'string',
            'int64': 'int',
            'float64': 'float',
            'datetime64[ns]': 'datetime',
            'bool': 'bool'
        }

    @staticmethod
    def clean_column_name(col):
        """Remove all forbidden characters from column names"""

        # Try to preserve accented characters
        cleaned_col = unicodedata.normalize('NFD', str(col)) \
            .encode('ascii', 'ignore') \
            .decode("utf-8")
        # Remove all non matching chars for Tableau WDC
        cleaned_col = re.sub(r'[^A-Za-z0-9_]+', '_', cleaned_col)
        return cleaned_col

    @staticmethod
    def replace_duplicated_column_name(cols):
        """Replace duplicated columns names"""
        cols_count_dict = dict(Counter(cols))
        # Filter unique items
        cols_count_dict = {key: value for (key, value) in cols_count_dict.items() if value > 1}
        unique_cols = list()
        for col in reversed(cols):
            idx = cols_count_dict.get(col, 0)
            unique_cols.insert(0, col if idx == 0 else col + '_' + str(idx))
            cols_count_dict[col] = idx - 1
        return unique_cols

    def get_schema(self, key):
        self.refresh(only_if_undefined=True)

        columns = [
            {
                'id': '.'.join(filter(None, key)) if isinstance(key, tuple) else key,
                'dataType':
                    self.types_mapping[str(value)] if str(value) in self.types_mapping else 'string'
            }
            for key, value in (self._prepare_dataframe()).dtypes.items()
        ]

        return {
            'id': key,
            'alias': self._alias,
            'columns': columns
        }

    def _prepare_dataframe(self, slice_from=None, slice_to=None):

        # Guarantee valid range for slicing
        if slice_from is None or slice_from < 0:
            slice_from = 0

        if slice_to is None:
            slice_to = len(self._dataframe)

        if slice_from > slice_to:
            raise IndexError(f"From ({slice_from}) can not be greater than To ({slice_to})")

        # Apply slicing to dataframe
        if slice_from < len(self._dataframe):
            # If slicing is in dataframe range
            output_df = self._dataframe.iloc[slice_from: min(slice_to, len(self._dataframe))]
        else:
            # If slicing is outside dataframe range, return an empty dataframe
            output_df = pd.DataFrame(columns=self._dataframe.columns)

        # Remove index if it is not required
        prep_df = output_df.reset_index() \
            if self._include_index \
            else output_df.reset_index(drop=True)

        # Flatten multi-index
        if isinstance(prep_df.columns, pd.MultiIndex):
            prep_df.columns = [self._index_separator.join(map(str, col)).strip()
                               for col in prep_df.columns.values]

        prep_df.columns = [DataFrameTable.clean_column_name(col) for col in prep_df.columns]
        prep_df.columns = DataFrameTable.replace_duplicated_column_name(prep_df.columns)

        return prep_df

    def refresh(self, only_if_undefined=False):
        # If DataFrame exists and it is not requested to update it then we do not need to refresh.
        # Otherwise if a refresh method has been set it is required to update the DataFrame.
        if (not only_if_undefined or self._dataframe is None) and self._refresh_method is not None:
            self._dataframe = self._refresh_method()

    def to_output(self, print_format='json', slice_from=None, slice_to=None):
        # Enforce print_format to be lower case
        print_format = print_format.lower()
        # Retrieve the DataFrame to be sent
        output_df = self._prepare_dataframe(slice_from=slice_from, slice_to=slice_to)

        # Directory of available formatter for output
        output_formatter = {
            'json': lambda df: df.to_json(orient='records', date_format="iso", date_unit="s")
        }

        if print_format in output_formatter.keys():
            return output_formatter[print_format](output_df)
        else:
            raise NotImplementedError(
                f"'{print_format}' format not supported."
                f"Please use one of the following : {output_formatter.keys()}"
            )



# Copyright (c) 2019 Capital Fund Management
# SPDX-License-Identifier: MIT

import json


class BaseTable:
    """
    Abstract class with default methods to be implemented, that represents a table-like object that
    can be used by Jupytab. DataFrames, SQL queries, CSV files... could be implemented as BaseTable
    child classes.
    """

    def __init__(self, alias):
        """
        alias -- A descriptive name of the table that will be displayed in Tableau.
        """
        self._alias = alias

    @property
    def alias(self):
        """
        A descriptive name for the Table.
        """
        return self._alias

    def get_schema(self, key):
        """
        Provide a table schema for Jupytab.

        Returns a dictionary that describes the table provided by the schema:
        {
            "id": :obj:`str`
            "alias": :obj:`str`
            "columns" :obj:`list`
        }

        key -- A uniquely identified schema.
        """
        raise NotImplementedError

    def refresh(self, only_if_undefined=False):
        """
        Refresh the table if an available refresh method is available.

        only_if_undefined -- true for refreshing only if there is no data currently available.
        """
        raise NotImplementedError

    def to_output(self, print_format='json', slice_from=None, slice_to=None):
        """
        Return the table contents as requested format, and selected slice
        """
        raise NotImplementedError


class Tables:
    """
    Table manager exposed as a dictionary to keep track of all registered tables and create a
    combined schema.
    """

    def __init__(self, *args):
        self.tables = {}

    def __setitem__(self, key, value):
        assert isinstance(value, BaseTable)
        self.tables[key.lower()] = value

    def __getitem__(self, key):
        return self.tables[key.lower()]

    def schema(self):
        """
        Return a list of all registered table schemas.
        """
        return [value.get_schema(key) for key, value in self.tables.items()]

    def render_schema(self, do_print=True):
        """
        Return the JSON with the table schemas or None (if it is printed [default]).

        To be used in the notebook in a single cell like the code provided below:

        ```
        # GET /schema
        tables.render_schema()
        ```

        This will generate a string in the cell output that Jupytab will be able to use.

        do_print -- If true, do not return the JSON object but print it instead.
        """
        rendered_schema = json.dumps(self.schema())
        if do_print:
            print(rendered_schema)
        else:
            return rendered_schema

    def render_data(self, request, do_print=True):
        """
        Return JSON with the table data or None (if it is printed [default]).

        To be used in the notebook in a single cell like the code provided below:

        ```
        # GET /data
        tables.render_data()
        ```

        This will generate a string in the cell output that Jupytab will be able to use.

        do_print -- If true, do not return the JSON object but print it instead.
        """

        request_dict = json.loads(request)['args']

        table_name = request_dict['table_name'][0]
        data_format = request_dict['format'][0] if 'format' in request_dict else 'json'
        slice_from = int(request_dict['from'][0]) if 'from' in request_dict else None
        slice_to = int(request_dict['to'][0]) if 'to' in request_dict else None
        refresh_required = request_dict['refresh'][0] == 'true'\
            if 'refresh' in request_dict else True

        table = self[table_name]

        if refresh_required:
            table.refresh()

        rendered_data = table.to_output(print_format=data_format,
                                        slice_from=slice_from,
                                        slice_to=slice_to)
        if do_print:
            print(rendered_data)
        else:
            return rendered_data



# Copyright (c) Capital Fund Management
# Distributed under the terms of the MIT License.

from setuptools import setup, find_packages

import distutils.cmd
import pathlib


class MakeVersionCommand(distutils.cmd.Command):
    """Prepare the version of jupytab"""

    description = "add extra file to make the version"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        version_file = pathlib.Path(__file__).parent.resolve()
        version_file = version_file.parent / 'VERSION'
        version = version_file.resolve().read_text().strip()
        version_py_file = pathlib.Path(__file__).parent / 'jupytab' / '__version__.py'
        with version_py_file.open('w') as stream:
            stream.write(f'__version__ = "{version}"\n')


def read_version():
    version = pathlib.Path(__file__).parent.resolve()
    version_master = version.parent / 'VERSION'
    version_master.resolve()
    version_4_release = version / 'jupytab' / '__version__.py'
    if version_4_release.exists():
        v = version_4_release.read_text().strip().split('=')[1].strip()
        # Let's remove quotes
        return v[1:len(v) - 1]
    elif version_master.exists():
        return version_master.read_text().strip()
    else:
        raise FileNotFoundError('no version file can be found')


README = """

Jupytab allows you to **explore in Tableau data which is generated
dynamically by a Jupyter Notebook**. You can thus create Tableau data
sources in a very flexible way using all the power of Python.
This is achieved by having Tableau access data through a **web
server created by Jupytab**.

This package is an helper package to split dependencies between
Jupytab server and your notebook. It only requires pandas.

The full documentation is available on the project's home page.
"""

# This call to setup() does all the work
setup(
    cmdclass=dict(make_version=MakeVersionCommand),
    name="jupytab",
    version=read_version(),
    description="Jupytab package to be used in notebooks",
    long_description_content_type="text/markdown",
    long_description=README,
    url="https://github.com/CFMTech/Jupytab",
    author="Brian Tribondeau",
    author_email="brian.tribondeau@cfm.fr",
    license="MIT",
    keywords="jupytab jupyter notebook wdc tableau",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=find_packages(exclude=["tests", "env", "docs", "build"]),
    python_requires='>=3.6',
    include_package_data=True,
    install_requires=["pandas"],
    project_urls={
        "Bug Tracker": "https://github.com/CFMTech/Jupytab/issues",
        "Documentation": "https://github.com/CFMTech/Jupytab",
        "Source Code": "https://github.com/CFMTech/Jupytab"
    }
)
